# General utils for foursight
from __future__ import print_function, unicode_literals
from datetime import datetime, timedelta
from dateutil import tz
import traceback
import signal
import time
import boto3
import sys
import os
import json
import re
from functools import wraps, partial
from .run_result import CheckResult, ActionResult, BadCheckOrAction
from .s3_connection import S3Connection
from .vars import FOURSIGHT_PREFIX

CHECK_DECO = 'check_function'
ACTION_DECO = 'action_function'
CHECK_TIMEOUT = 870  # in seconds. set to less than lambda limit (900 s)
POLL_INTERVAL = 10  # check child process every 10 seconds

# compare strings in both python 2 and python 3
# in other files, compare with utils.basestring
try:
    basestring = basestring
except NameError:
    basestring = str


def get_stage_info():
    """
    Returns a dictionary with stage info and queue/check runner names that
    depend on that. If test=True, use the test runner_name
    """
    # set environmental variables in .chalice/config.json
    stage = os.environ.get('chalice_stage', 'dev') # default to dev
    queue_name = '-'.join([FOURSIGHT_PREFIX, stage, 'check_queue'])
    # when testing, use dev stage with test queue
    if stage == 'test':
        stage = 'dev'
        queue_name = FOURSIGHT_PREFIX + '-test-check_queue'
    runner_name = '-'.join([FOURSIGHT_PREFIX, stage, 'check_runner'])
    return {'stage': stage, 'queue_name': queue_name, 'runner_name': runner_name}


def list_environments():
    """
    Lists all environments in the foursight-envs s3. Returns a list of names
    """
    s3_connection = S3Connection(FOURSIGHT_PREFIX + '-envs')
    return s3_connection.list_all_keys()


def init_check_res(connection, name, init_uuid=None):
    """
    Initialize a CheckResult object, which holds all information for a
    check and methods necessary to store and retrieve latest/historical
    results. name is the only required parameter and MUST be equal to
    the method name of the check as defined in CheckSuite.

    init_uuid is a a result uuid that the check will look for upon initialization.
    If found, the check fields will be pre-populated with its results.

    This function will be deprecated in an upcoming version.
    """
    return CheckResult(connection, name, init_uuid=init_uuid)


def init_action_res(connection, name):
    """
    Similar to init_check_res, but meant to be used for ActionResult items

    This function will be deprecated in an upcoming version.
    """
    return ActionResult(connection, name)


def get_methods_by_deco(cls, decorator):
    """
    Returns all methods in cls/module with decorator as a list;
    the decorator is set in check_function()
    """
    methods = []
    for maybeDecorated in cls.__dict__.values():
        if hasattr(maybeDecorated, 'check_decorator'):
            if maybeDecorated.check_decorator == decorator:
                methods.append(maybeDecorated)
    return methods


def check_method_deco(method, decorator):
    """
    See if the given method has the given decorator. Returns True if so,
    False if not.
    """
    return hasattr(method, 'check_decorator') and method.check_decorator == decorator


def handle_kwargs(kwargs, default_kwargs):
    # add all default args that are not defined in kwargs
    # also ensure 'uuid' and 'primary' are in there
    for key in default_kwargs:
        if key not in kwargs:
            kwargs[key] = default_kwargs[key]
    if 'uuid' not in kwargs:
        kwargs['uuid'] = datetime.utcnow().isoformat()
    if 'primary' not in kwargs:
        kwargs['primary'] = False
    return kwargs


def pid_is_alive(pid):
    """
    Returns True if pid is still alive
    """
    try:
        os.kill(pid, 0)  # do not send a signal, just error check
    except OSError:
        return False
    else:
        return True


def check_function(*default_args, **default_kwargs):
    """
    Import decorator, used to decorate all checks.
    Sets the check_decorator attribute so that methods can be fetched.
    Any kwargs provided to the decorator will be passed to the function
    if no kwargs are explicitly passed.
    Handles all exceptions within running of the check, including validation
    issues/some common errors when writing checks. Will also keep track of overall
    runtime and cancel the check with status=ERROR if runtime exceeds CHECK_TIMEOUT.
    If an exception is raised, will store the result in full_output and
    return an ERROR CheckResult.
    """
    def check_deco(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            start_time = time.time()
            kwargs = handle_kwargs(kwargs, default_kwargs)
            parent_pid = os.getpid()
            child_pid = os.fork()
            if child_pid != 0:  # we are the parent who will execute the check
                try:
                    check = func(*args, **kwargs)
                    check.validate()
                except Exception as e:
                    # connection should be the first (and only) positional arg
                    check = CheckResult(args[0], func.__name__)
                    check.status = 'ERROR'
                    check.description = 'Check failed to run. See full output.'
                    check.full_output = traceback.format_exc().split('\n')
                kwargs['runtime_seconds'] = round(time.time() - start_time, 2)
                check.kwargs = kwargs
                os.kill(child_pid, signal.SIGKILL)  # we finished, so kill child
                return check.store_result()
            else:  # we are the child who handles the timeout
                partials = {'name': func.__name__, 'kwargs': kwargs, 'is_check': True,
                            'start_time': start_time, 'connection': args[0]}
                do_timeout(parent_pid, partials)

        wrapper.check_decorator = CHECK_DECO
        return wrapper
    return check_deco


def action_function(*default_args, **default_kwargs):
    """
    Import decorator, used to decorate all actions.
    Required for action functions.
    Any kwargs provided to the decorator will be passed to the function
    if no kwargs are explicitly passed.
    Handles all exceptions within running of the action, including validation
    issues/some common errors when writing actions. Will also keep track of overall
    runtime and cancel the check with status=ERROR if runtime exceeds CHECK_TIMEOUT.
    If an exception is raised, will store the result in output and return an
    ActionResult with status FAIL.
    """
    def action_deco(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            start_time = time.time()
            kwargs = handle_kwargs(kwargs, default_kwargs)
            parent_pid = os.getpid()
            child_pid = os.fork()
            if child_pid != 0:  # we are the parent who will execute the check
                try:
                    if 'check_name' not in kwargs or 'called_by' not in kwargs:
                        raise BadCheckOrAction('Action requires check_name and called_by in its kwargs.')
                    action = func(*args, **kwargs)
                    action.validate()
                except Exception as e:
                    # connection should be the first (and only) positional arg
                    action = ActionResult(args[0], func.__name__)
                    action.status = 'FAIL'
                    action.description = 'Action failed to run. See output.'
                    action.output = traceback.format_exc().split('\n')
                kwargs['runtime_seconds'] = round(time.time() - start_time, 2)
                action.kwargs = kwargs
                os.kill(child_pid, signal.SIGKILL)  # we finished, so kill child
                return action.store_result()
            else:  # we are the child who handles the timeout
                partials = {'name': func.__name__, 'kwargs': kwargs, 'is_check': False,
                            'start_time': start_time, 'connection': args[0]}
                do_timeout(parent_pid, partials)

        wrapper.check_decorator = ACTION_DECO
        return wrapper
    return action_deco


class BadCheckSetup(Exception):
    """
    Generic exception for an issue with a check setup.
    __init__ takes some string error message
    """
    def __init__(self, message=None):
        # default error message if none provided
        if message is None:
            message = "Malformed check setup found."
        super().__init__(message)


def do_timeout(parent_pid, partials):
    """ Wrapper for below method that handles:
            1. Polling across the CHECK_TIMEOUT at POLL_INTERVAL
            2. Exiting if we succeeded (the parent process died)
            3. Killing the parent if it timed out
            4. Invoking the timeout handler if it timed out

        :arg parent_pid: parent pid to check on/kill if necessary
        :arg partials: partial result to be passed to timeout handler if necessary
    """
    for t in range(CHECK_TIMEOUT // POLL_INTERVAL):  # Divide CHECK_TIMEOUT into POLL_INTERVAL slices
        time.sleep(POLL_INTERVAL)
        if not pid_is_alive(parent_pid):
            sys.exit(0)

    # We have timed out. Kill the parent and invoke the timeout handler.
    # NOTE: Timeouts in Pytest will trigger undefined behavior since the parent is Pytest, not the thing
    # executing the check. Execute Pytest with --forked option to override this.
    os.kill(parent_pid, signal.SIGTERM)
    timeout_handler(partials)


def timeout_handler(partials, signum=None, frame=None):
    """
    Custom handler for signal that stores the current check
    or action with the appropriate information and then exits using sys.exit
    """
    if partials['is_check']:
        result = CheckResult(partials['connection'], partials['name'])
        result.status = 'ERROR'
    else:
        result = ActionResult(partials['connection'], partials['name'])
        result.status = 'FAIL'
    result.description = 'AWS lambda execution reached the time limit. Please see check/action code.'
    kwargs = partials['kwargs']
    kwargs['runtime_seconds'] = round(time.time() - partials['start_time'], 2)
    result.kwargs = kwargs
    result.store_result()
    # need to delete the sqs message and propogate if this is using the queue
    if kwargs.get('_run_info') and {'receipt', 'sqs_url'} <= set(kwargs['_run_info'].keys()):
        runner_input = {'sqs_url': kwargs['_run_info']['sqs_url']}
        delete_message_and_propogate(runner_input, kwargs['_run_info']['receipt'])
    sys.exit('-RUN-> TIMEOUT for execution of %s. Elapsed time is %s seconds; keep under %s.'
          % (partials['name'], kwargs['runtime_seconds'], CHECK_TIMEOUT))


def parse_datetime_to_utc(time_str, manual_format=None):
    """
    Attempt to parse the string time_str with the given string format.
    If no format is given, attempt to automatically parse the given string
    that may or may not contain timezone information.
    Returns a datetime object of the string in UTC
    or None if the parsing was unsuccessful.
    """
    if manual_format and isinstance(manual_format, basestring):
        timeobj = datetime.strptime(time_str, manual_format)
    else:  # automatic parsing
        if len(time_str) > 26 and time_str[26] in ['+', '-']:
            try:
                timeobj = datetime.strptime(time_str[:26],'%Y-%m-%dT%H:%M:%S.%f')
            except ValueError:
                return None
            if time_str[26]=='+':
                timeobj -= timedelta(hours=int(time_str[27:29]), minutes=int(time_str[30:]))
            elif time_str[26]=='-':
                timeobj += timedelta(hours=int(time_str[27:29]), minutes=int(time_str[30:]))
        elif len(time_str) == 26 and '+' not in time_str[-6:] and '-' not in time_str[-6:]:
            # nothing known about tz, just parse it without tz in this cause
            try:
                timeobj = datetime.strptime(time_str[0:26],'%Y-%m-%dT%H:%M:%S.%f')
            except ValueError:
                return None
        else:
            # last try: attempt without milliseconds
            try:
                timeobj = datetime.strptime(time_str, "%Y-%m-%dT%H:%M:%S")
            except ValueError:
                return None
    return timeobj.replace(tzinfo=tz.tzutc())


def convert_camel_to_snake(name):
    """
    Convert a given CamelCase string to snake_case
    Credit: https://stackoverflow.com/questions/1175208/elegant-python-function-to-convert-camelcase-to-snake-case
    """
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


##### SQS utils #####


def invoke_check_runner(runner_input):
    """
    Simple function to invoke the next check_runner lambda with runner_input
    (dict containing {'sqs_url': <str>})
    """
    client = boto3.client('lambda')
    # InvocationType='Event' makes asynchronous
    # try/except while async invokes are problematic
    try:
        response = client.invoke(
            FunctionName=get_stage_info()['runner_name'],
            InvocationType='Event',
            Payload=json.dumps(runner_input)
        )
    except:
        response = client.invoke(
            FunctionName=get_stage_info()['runner_name'],
            Payload=json.dumps(runner_input)
        )
    return response


def delete_message_and_propogate(runner_input, receipt, propogate=True):
    """
    Delete the message with given receipt from sqs queue and invoke the next
    lambda runner.

    Args:
        runner_input (dict): runner info, should minimally have 'sqs_url'
        receipt (str): SQS message receipt
        propogate (bool): if True (default), invoke another check runner lambda

    Returns:
        None
    """
    sqs_url = runner_input.get('sqs_url')
    if not sqs_url or not receipt:
        return
    client = boto3.client('sqs')
    client.delete_message(
        QueueUrl=sqs_url,
        ReceiptHandle=receipt
    )
    if propogate is True:
        invoke_check_runner(runner_input)


def recover_message_and_propogate(runner_input, receipt, propogate=True):
    """
    Recover the message with given receipt to sqs queue and invoke the next
    lambda runner.

    Changing message VisibilityTimeout to 15 seconds means the message will be
    available to the queue in that much time. This is a slight lag to allow
    dependencies to process.
    NOTE: VisibilityTimeout should be less than WaitTimeSeconds in run_check_runner

    Args:
        runner_input (dict): runner info, should minimally have 'sqs_url'
        receipt (str): SQS message receipt
        propogate (bool): if True (default), invoke another check runner lambda

    Returns:
        None
    """
    sqs_url = runner_input.get('sqs_url')
    if not sqs_url or not receipt:
        return
    client = boto3.client('sqs')
    client.change_message_visibility(
        QueueUrl=sqs_url,
        ReceiptHandle=receipt,
        VisibilityTimeout=15
    )
    if propogate is True:
        invoke_check_runner(runner_input)


def get_sqs_queue():
    """
    Returns boto3 sqs resource
    """
    queue_name = get_stage_info()['queue_name']
    sqs = boto3.resource('sqs')
    try:
        queue = sqs.get_queue_by_name(QueueName=queue_name)
    except:
        queue = sqs.create_queue(
            QueueName=queue_name,
            Attributes={
                'VisibilityTimeout': '300',
                'MessageRetentionPeriod': '3600'
            }
        )
    return queue


def collect_run_info(run_uuid):
    """
    Returns a set of run checks under this run uuid
    """
    s3_connection = S3Connection(FOURSIGHT_PREFIX + '-runs')
    run_prefix = ''.join([run_uuid, '/'])
    complete = s3_connection.list_all_keys_w_prefix(run_prefix)
    # eliminate duplicates
    return set(complete)


def send_sqs_messages(queue, environ, check_vals, uuid=None):
    """
    Send messages to SQS queue. Check_vals are entries within a check_group.
    Optionally, provide a uuid that will be queued as the uuid for the run; if
    not provided, datetime.utcnow is used

    Args:
        queue: boto3 sqs resource (from get_sqs_queue)
        environ (str): foursight environment name
        check_vals (list): list of formatted check vals, like those from
            check_utils.get_check_schedule
        uuid (str): optional string uuid

    Returns:
        str: uuid of queued messages
    """
    # uuid used as the MessageGroupId
    if not uuid:
        uuid = datetime.utcnow().isoformat()
    # append environ and uuid as first elements to all check_vals
    proc_vals = [[environ, uuid] + val for val in check_vals]
    for val in proc_vals:
        response = queue.send_message(MessageBody=json.dumps(val))
    return uuid


def get_sqs_attributes(sqs_url):
    """
    Returns a dict of the desired attributes form the queue with given url
    """
    backup = {
        'ApproximateNumberOfMessages': 'ERROR',
        'ApproximateNumberOfMessagesNotVisible': 'ERROR'
    }
    client = boto3.client('sqs')
    try:
        result = client.get_queue_attributes(
            QueueUrl=sqs_url,
            AttributeNames=[
                'ApproximateNumberOfMessages',
                'ApproximateNumberOfMessagesNotVisible'
            ]
        )
    except:
        return backup
    return result.get('Attributes', backup)


def load_json(rel, fname):
    """ Loads json file fname from rel/fname """
    path = os.path.join(os.path.dirname(rel), fname)
    with open(path, 'r') as f:
        return json.load(f)


def store_json(rel, fname, data):
    """ Stores json data into fname at rel/fname """
    path = os.path.join(os.path.dirname(rel), fname)
    with open(path, 'w+') as f:
        return json.dump(data, f)


def cat_indices(client):
    """ Wrapper function for the ES API _cat/indices so that the result returned is comprehensible.

        :param client: es client to use
        :returns: 2-tuple lists of header, rows
    """
    if not client:
        return [], []
    indices = client.cat.indices(v=True).split('\n')
    split_indices = [ind.split() for ind in indices]
    headers = split_indices.pop(0)  # first row is header
    return headers, split_indices

